@@ -1,4 +1,4 @@
-# Copyright 2012-2021 The Meson development team
+# Copyright 2012-2022 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -11,9 +11,10 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import annotations
 
 from ..mesonlib import (
-    MachineChoice, MesonException, EnvironmentException,
+    MesonException, EnvironmentException, join_args,
     search_version, is_windows, Popen_safe, windows_proof_rm,
 )
 from ..envconfig import BinaryTable
@@ -23,6 +24,7 @@ from ..linkers import (
     guess_win_linker,
     guess_nix_linker,
     AIXArLinker,
+    AppleArLinker,
     ArLinker,
     ArmarLinker,
     ArmClangDynamicLinker,
@@ -40,7 +42,6 @@ from ..linkers import (
     NvidiaHPC_DynamicLinker,
     PGIDynamicLinker,
     PGIStaticLinker,
-    StaticLinker,
     Xc16Linker,
     Xc16DynamicLinker,
     XilinkDynamicLinker,
@@ -50,9 +51,7 @@ from ..linkers import (
     VisualStudioLikeLinkerMixin,
     WASMDynamicLinker,
 )
-from .compilers import Compiler
 from .c import (
-    CCompiler,
     AppleClangCCompiler,
     ArmCCompiler,
     ArmclangCCompiler,
@@ -74,7 +73,6 @@ from .c import (
     VisualStudioCCompiler,
 )
 from .cpp import (
-    CPPCompiler,
     AppleClangCPPCompiler,
     ArmCPPCompiler,
     ArmclangCPPCompiler,
@@ -102,7 +100,6 @@ from .d import (
 )
 from .cuda import CudaCompiler
 from .fortran import (
-    FortranCompiler,
     ArmLtdFlangFortranCompiler,
     G95FortranCompiler,
     GnuFortranCompiler,
@@ -119,13 +116,11 @@ from .fortran import (
 )
 from .java import JavaCompiler
 from .objc import (
-    ObjCCompiler,
     AppleClangObjCCompiler,
     ClangObjCCompiler,
     GnuObjCCompiler,
 )
 from .objcpp import (
-    ObjCPPCompiler,
     AppleClangObjCPPCompiler,
     ClangObjCPPCompiler,
     GnuObjCPPCompiler,
@@ -147,7 +142,15 @@ import os
 import typing as T
 
 if T.TYPE_CHECKING:
+    from .compilers import Compiler
+    from .c import CCompiler
+    from .cpp import CPPCompiler
+    from .fortran import FortranCompiler
+    from .objc import ObjCCompiler
+    from .objcpp import ObjCPPCompiler
+    from ..linkers import StaticLinker
     from ..environment import Environment
+    from ..mesonlib import MachineChoice
     from ..programs import ExternalProgram
 
 
@@ -260,7 +263,7 @@ def _handle_exceptions(
     if exceptions:
         errmsg += '\nThe following exception(s) were encountered:'
         for c, e in exceptions.items():
-            errmsg += f'\nRunning "{c}" gave "{e}"'
+            errmsg += f'\nRunning `{c}` gave "{e}"'
     raise EnvironmentException(errmsg)
 
 
@@ -309,7 +312,7 @@ def detect_static_linker(env: 'Environment', compiler: Compiler) -> StaticLinker
         try:
             p, out, err = Popen_safe(linker + [arg])
         except OSError as e:
-            popen_exceptions[' '.join(linker + [arg])] = e
+            popen_exceptions[join_args(linker + [arg])] = e
             continue
         if "xilib: executing 'lib'" in err:
             return IntelVisualStudioLinker(linker, getattr(compiler, 'machine', None))
@@ -340,13 +343,13 @@ def detect_static_linker(env: 'Environment', compiler: Compiler) -> StaticLinker
         if out.startswith('The CompCert'):
             return CompCertLinker(linker)
         if p.returncode == 0:
-            return ArLinker(linker)
+            return ArLinker(compiler.for_machine, linker)
         if p.returncode == 1 and err.startswith('usage'): # OSX
-            return ArLinker(linker)
+            return AppleArLinker(compiler.for_machine, linker)
         if p.returncode == 1 and err.startswith('Usage'): # AIX
             return AIXArLinker(linker)
         if p.returncode == 1 and err.startswith('ar: bad option: --'): # Solaris
-            return ArLinker(linker)
+            return ArLinker(compiler.for_machine, linker)
     _handle_exceptions(popen_exceptions, linkers, 'linker')
 
 
@@ -414,10 +417,16 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
         else:
             arg = '--version'
 
+        cmd = compiler + [arg]
         try:
-            p, out, err = Popen_safe(compiler + [arg])
+            mlog.debug('-----')
+            mlog.debug(f'Detecting compiler via: {join_args(cmd)}')
+            p, out, err = Popen_safe(cmd)
+            mlog.debug(f'compiler returned {p}')
+            mlog.debug(f'compiler stdout:\n{out}')
+            mlog.debug(f'compiler stderr:\n{err}')
         except OSError as e:
-            popen_exceptions[' '.join(compiler + [arg])] = e
+            popen_exceptions[join_args(cmd)] = e
             continue
 
         if 'ccrx' in compiler_name:
@@ -438,7 +447,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
         if guess_gcc_or_lcc:
             defines = _get_gnu_compiler_defines(compiler)
             if not defines:
-                popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
+                popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                 continue
 
             if guess_gcc_or_lcc == 'lcc':
@@ -448,7 +457,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 version = _get_gnu_version_from_defines(defines)
                 cls = GnuCCompiler if lang == 'c' else GnuCPPCompiler
 
-            linker = guess_nix_linker(env, compiler, cls, for_machine)
+            linker = guess_nix_linker(env, compiler, cls, version, for_machine)
 
             return cls(
                 ccache + compiler, version, for_machine, is_cross,
@@ -475,16 +484,14 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 exe_wrap, linker=linker, full_version=full_version)
 
         if 'Arm C/C++/Fortran Compiler' in out:
-            arm_ver_match = re.search(r'version (\d+)\.(\d+) \(build number (\d+)\)', out)
-            arm_ver_major = arm_ver_match.group(1)
-            arm_ver_minor = arm_ver_match.group(2)
-            arm_ver_build = arm_ver_match.group(3)
-            version = '.'.join([arm_ver_major, arm_ver_minor, arm_ver_build])
+            arm_ver_match = re.search(r'version (\d+)\.(\d+)\.?(\d+)? \(build number (\d+)\)', out)
+            assert arm_ver_match is not None, 'for mypy'  # because mypy *should* be complaning that this could be None
+            version = '.'.join([x for x in arm_ver_match.groups() if x is not None])
             if lang == 'c':
                 cls = ArmLtdClangCCompiler
             elif lang == 'cpp':
                 cls = ArmLtdClangCPPCompiler
-            linker = guess_nix_linker(env, compiler, cls, for_machine)
+            linker = guess_nix_linker(env, compiler, cls, version, for_machine)
             return cls(
                 ccache + compiler, version, for_machine, is_cross, info,
                 exe_wrap, linker=linker)
@@ -496,7 +503,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
             # output format in future versions
             arm_ver_match = re.search('.*Component.*', out)
             if arm_ver_match is None:
-                popen_exceptions[' '.join(compiler)] = 'version string not found'
+                popen_exceptions[join_args(compiler)] = 'version string not found'
                 continue
             arm_ver_str = arm_ver_match.group(0)
             # Override previous values
@@ -515,7 +522,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
             try:
                 p, out, err = Popen_safe(compiler + [arg])
             except OSError as e:
-                popen_exceptions[' '.join(compiler + [arg])] = e
+                popen_exceptions[join_args(compiler + [arg])] = e
             version = search_version(out)
             match = re.search('^Target: (.*?)-', out, re.MULTILINE)
             if match:
@@ -523,7 +530,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
             else:
                 target = 'unknown target'
             cls = ClangClCCompiler if lang == 'c' else ClangClCPPCompiler
-            linker = guess_win_linker(env, ['lld-link'], cls, for_machine)
+            linker = guess_win_linker(env, ['lld-link'], cls, version, for_machine)
             return cls(
                 compiler, version, for_machine, is_cross, info, target,
                 exe_wrap, linker=linker)
@@ -544,11 +551,11 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 # style ld, but for clang on "real" windows we'll use
                 # either link.exe or lld-link.exe
                 try:
-                    linker = guess_win_linker(env, compiler, cls, for_machine, invoked_directly=False)
+                    linker = guess_win_linker(env, compiler, cls, version, for_machine, invoked_directly=False)
                 except MesonException:
                     pass
             if linker is None:
-                linker = guess_nix_linker(env, compiler, cls, for_machine)
+                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
 
             return cls(
                 ccache + compiler, version, for_machine, is_cross, info,
@@ -582,7 +589,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 m = f'Failed to detect MSVC compiler target architecture: \'cl /?\' output is\n{cl_signature}'
                 raise EnvironmentException(m)
             cls = VisualStudioCCompiler if lang == 'c' else VisualStudioCPPCompiler
-            linker = guess_win_linker(env, ['link'], cls, for_machine)
+            linker = guess_win_linker(env, ['link'], cls, version, for_machine)
             # As of this writing, CCache does not support MSVC but sccache does.
             if 'sccache' in ccache:
                 final_compiler = ccache + compiler
@@ -607,7 +614,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 info, exe_wrap, linker=linker)
         if '(ICC)' in out:
             cls = IntelCCompiler if lang == 'c' else IntelCPPCompiler
-            l = guess_nix_linker(env, compiler, cls, for_machine)
+            l = guess_nix_linker(env, compiler, cls, version, for_machine)
             return cls(
                 ccache + compiler, version, for_machine, is_cross, info,
                 exe_wrap, full_version=full_version, linker=l)
@@ -675,7 +682,7 @@ def detect_cuda_compiler(env: 'Environment', for_machine: MachineChoice) -> Comp
         try:
             p, out, err = Popen_safe(compiler + [arg])
         except OSError as e:
-            popen_exceptions[' '.join(compiler + [arg])] = e
+            popen_exceptions[join_args(compiler + [arg])] = e
             continue
         # Example nvcc printout:
         #
@@ -711,7 +718,7 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
             try:
                 p, out, err = Popen_safe(compiler + [arg])
             except OSError as e:
-                popen_exceptions[' '.join(compiler + [arg])] = e
+                popen_exceptions[join_args(compiler + [arg])] = e
                 continue
 
             version = search_version(out)
@@ -726,37 +733,35 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
             if guess_gcc_or_lcc:
                 defines = _get_gnu_compiler_defines(compiler)
                 if not defines:
-                    popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
+                    popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                     continue
                 if guess_gcc_or_lcc == 'lcc':
                     version = _get_lcc_version_from_defines(defines)
                     cls = ElbrusFortranCompiler
-                    linker = guess_nix_linker(env, compiler, cls, for_machine)
+                    linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                     return cls(
                         compiler, version, for_machine, is_cross, info,
                         exe_wrap, defines, full_version=full_version, linker=linker)
                 else:
                     version = _get_gnu_version_from_defines(defines)
                     cls = GnuFortranCompiler
-                    linker = guess_nix_linker(env, compiler, cls, for_machine)
+                    linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                     return cls(
                         compiler, version, for_machine, is_cross, info,
                         exe_wrap, defines, full_version=full_version, linker=linker)
 
             if 'Arm C/C++/Fortran Compiler' in out:
                 cls = ArmLtdFlangFortranCompiler
-                arm_ver_match = re.search(r'version (\d+)\.(\d+) \(build number (\d+)\)', out)
-                arm_ver_major = arm_ver_match.group(1)
-                arm_ver_minor = arm_ver_match.group(2)
-                arm_ver_build = arm_ver_match.group(3)
-                version = '.'.join([arm_ver_major, arm_ver_minor, arm_ver_build])
-                linker = guess_nix_linker(env, compiler, cls, for_machine)
+                arm_ver_match = re.search(r'version (\d+)\.(\d+)\.?(\d+)? \(build number (\d+)\)', out)
+                assert arm_ver_match is not None, 'for mypy'  # because mypy *should* be complaning that this could be None
+                version = '.'.join([x for x in arm_ver_match.groups() if x is not None])
+                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                 return cls(
                     ccache + compiler, version, for_machine, is_cross, info,
                     exe_wrap, linker=linker)
             if 'G95' in out:
                 cls = G95FortranCompiler
-                linker = guess_nix_linker(env, compiler, cls, for_machine)
+                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                 return G95FortranCompiler(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
@@ -764,7 +769,7 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
             if 'Sun Fortran' in err:
                 version = search_version(err)
                 cls = SunFortranCompiler
-                linker = guess_nix_linker(env, compiler, cls, for_machine)
+                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                 return SunFortranCompiler(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
@@ -780,7 +785,7 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
                     target, exe_wrap, linker=linker)
 
             if 'ifort (IFORT)' in out:
-                linker = guess_nix_linker(env, compiler, IntelFortranCompiler, for_machine)
+                linker = guess_nix_linker(env, compiler, IntelFortranCompiler, version, for_machine)
                 return IntelFortranCompiler(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
@@ -810,14 +815,14 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
 
             if 'flang' in out or 'clang' in out:
                 linker = guess_nix_linker(env,
-                                          compiler, FlangFortranCompiler, for_machine)
+                                          compiler, FlangFortranCompiler, version, for_machine)
                 return FlangFortranCompiler(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
 
             if 'Open64 Compiler Suite' in err:
                 linker = guess_nix_linker(env,
-                                          compiler, Open64FortranCompiler, for_machine)
+                                          compiler, Open64FortranCompiler, version, for_machine)
                 return Open64FortranCompiler(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
@@ -855,17 +860,17 @@ def _detect_objc_or_objcpp_compiler(env: 'Environment', for_machine: MachineChoi
         try:
             p, out, err = Popen_safe(compiler + arg)
         except OSError as e:
-            popen_exceptions[' '.join(compiler + arg)] = e
+            popen_exceptions[join_args(compiler + arg)] = e
             continue
         version = search_version(out)
         if 'Free Software Foundation' in out:
             defines = _get_gnu_compiler_defines(compiler)
             if not defines:
-                popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
+                popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                 continue
             version = _get_gnu_version_from_defines(defines)
             comp = GnuObjCCompiler if objc else GnuObjCPPCompiler
-            linker = guess_nix_linker(env, compiler, comp, for_machine)
+            linker = guess_nix_linker(env, compiler, comp, version, for_machine)
             return comp(
                 ccache + compiler, version, for_machine, is_cross, info,
                 exe_wrap, defines, linker=linker)
@@ -873,7 +878,7 @@ def _detect_objc_or_objcpp_compiler(env: 'Environment', for_machine: MachineChoi
             linker = None
             defines = _get_clang_compiler_defines(compiler)
             if not defines:
-                popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
+                popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                 continue
             if 'Apple' in out:
                 comp = AppleClangObjCCompiler if objc else AppleClangObjCPPCompiler
@@ -882,12 +887,12 @@ def _detect_objc_or_objcpp_compiler(env: 'Environment', for_machine: MachineChoi
             if 'windows' in out or env.machines[for_machine].is_windows():
                 # If we're in a MINGW context this actually will use a gnu style ld
                 try:
-                    linker = guess_win_linker(env, compiler, comp, for_machine)
+                    linker = guess_win_linker(env, compiler, comp, version, for_machine)
                 except MesonException:
                     pass
 
             if not linker:
-                linker = guess_nix_linker(env, compiler, comp, for_machine)
+                linker = guess_nix_linker(env, compiler, comp, version, for_machine)
             return comp(
                 ccache + compiler, version, for_machine,
                 is_cross, info, exe_wrap, linker=linker, defines=defines)
@@ -904,7 +909,7 @@ def detect_java_compiler(env: 'Environment', for_machine: MachineChoice) -> Comp
     try:
         p, out, err = Popen_safe(exelist + ['-version'])
     except OSError:
-        raise EnvironmentException('Could not execute Java compiler "{}"'.format(' '.join(exelist)))
+        raise EnvironmentException('Could not execute Java compiler: {}'.format(join_args(exelist)))
     if 'javac' in out or 'javac' in err:
         version = search_version(err if 'javac' in err else out)
         if not version or version == 'unknown version':
@@ -914,7 +919,7 @@ def detect_java_compiler(env: 'Environment', for_machine: MachineChoice) -> Comp
         comp_class = JavaCompiler
         env.coredata.add_lang_args(comp_class.language, comp_class, for_machine, env)
         return comp_class(exelist, version, for_machine, info)
-    raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
+    raise EnvironmentException('Unknown compiler: ' + join_args(exelist))
 
 def detect_cs_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
     compilers, ccache, exe_wrap = _get_compilers(env, 'cs', for_machine)
@@ -924,7 +929,7 @@ def detect_cs_compiler(env: 'Environment', for_machine: MachineChoice) -> Compil
         try:
             p, out, err = Popen_safe(comp + ['--version'])
         except OSError as e:
-            popen_exceptions[' '.join(comp + ['--version'])] = e
+            popen_exceptions[join_args(comp + ['--version'])] = e
             continue
 
         version = search_version(out)
@@ -952,7 +957,7 @@ def detect_cython_compiler(env: 'Environment', for_machine: MachineChoice) -> Co
         try:
             err = Popen_safe(comp + ['-V'])[2]
         except OSError as e:
-            popen_exceptions[' '.join(comp + ['-V'])] = e
+            popen_exceptions[join_args(comp + ['-V'])] = e
             continue
 
         version = search_version(err)
@@ -974,13 +979,13 @@ def detect_vala_compiler(env: 'Environment', for_machine: MachineChoice) -> Comp
     try:
         p, out = Popen_safe(exelist + ['--version'])[0:2]
     except OSError:
-        raise EnvironmentException('Could not execute Vala compiler "{}"'.format(' '.join(exelist)))
+        raise EnvironmentException('Could not execute Vala compiler: {}'.format(join_args(exelist)))
     version = search_version(out)
     if 'Vala' in out:
         comp_class = ValaCompiler
         env.coredata.add_lang_args(comp_class.language, comp_class, for_machine, env)
         return comp_class(exelist, version, for_machine, is_cross, info)
-    raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
+    raise EnvironmentException('Unknown compiler: ' + join_args(exelist))
 
 def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> RustCompiler:
     popen_exceptions = {}  # type: T.Dict[str, Exception]
@@ -997,7 +1002,7 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
         try:
             out = Popen_safe(compiler + arg)[1]
         except OSError as e:
-            popen_exceptions[' '.join(compiler + arg)] = e
+            popen_exceptions[join_args(compiler + arg)] = e
             continue
 
         version = search_version(out)
@@ -1032,7 +1037,7 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                 extra_args: T.Dict[str, T.Union[str, bool]] = {}
                 always_args: T.List[str] = []
                 if is_link_exe:
-                    compiler.extend(cls.use_linker_args(cc.linker.exelist[0]))
+                    compiler.extend(cls.use_linker_args(cc.linker.exelist[0], ''))
                     extra_args['direct'] = True
                     extra_args['machine'] = cc.linker.machine
                 else:
@@ -1040,7 +1045,7 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                     if 'ccache' in exelist[0]:
                         del exelist[0]
                     c = exelist.pop(0)
-                    compiler.extend(cls.use_linker_args(c))
+                    compiler.extend(cls.use_linker_args(c, ''))
 
                     # Also ensure that we pass any extra arguments to the linker
                     for l in exelist:
@@ -1049,6 +1054,7 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                 # This trickery with type() gets us the class of the linker
                 # so we can initialize a new copy for the Rust Compiler
                 # TODO rewrite this without type: ignore
+                assert cc.linker is not None, 'for mypy'
                 if is_link_exe:
                     linker = type(cc.linker)(for_machine, always_args, exelist=cc.linker.exelist,   # type: ignore
                                              version=cc.linker.version, **extra_args)               # type: ignore
@@ -1058,12 +1064,12 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                                              **extra_args)
             elif 'link' in override[0]:
                 linker = guess_win_linker(env,
-                                          override, cls, for_machine, use_linker_prefix=False)
+                                          override, cls, version, for_machine, use_linker_prefix=False)
                 # rustc takes linker arguments without a prefix, and
                 # inserts the correct prefix itself.
                 assert isinstance(linker, VisualStudioLikeLinkerMixin)
                 linker.direct = True
-                compiler.extend(cls.use_linker_args(linker.exelist[0]))
+                compiler.extend(cls.use_linker_args(linker.exelist[0], ''))
             else:
                 # On linux and macos rust will invoke the c compiler for
                 # linking, on windows it will use lld-link or link.exe.
@@ -1075,7 +1081,7 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                 # Of course, we're not going to use any of that, we just
                 # need it to get the proper arguments to pass to rustc
                 c = linker.exelist[1] if linker.exelist[0].endswith('ccache') else linker.exelist[0]
-                compiler.extend(cls.use_linker_args(c))
+                compiler.extend(cls.use_linker_args(c, ''))
 
             env.coredata.add_lang_args(cls.language, cls, for_machine, env)
             return cls(
@@ -1116,7 +1122,7 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
         try:
             p, out = Popen_safe(exelist + ['--version'])[0:2]
         except OSError as e:
-            popen_exceptions[' '.join(exelist + ['--version'])] = e
+            popen_exceptions[join_args(exelist + ['--version'])] = e
             continue
         version = search_version(out)
         full_version = out.split('\n', 1)[0]
@@ -1134,7 +1140,7 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
                     objfile = os.path.basename(f)[:-1] + 'obj'
                     linker = guess_win_linker(env,
                                               exelist,
-                                              LLVMDCompiler, for_machine,
+                                              LLVMDCompiler, full_version, for_machine,
                                               use_linker_prefix=True, invoked_directly=False,
                                               extra_args=[f])
                 else:
@@ -1142,7 +1148,7 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
                     # Clean it up.
                     objfile = os.path.basename(f)[:-1] + 'o'
                     linker = guess_nix_linker(env,
-                                              exelist, LLVMDCompiler, for_machine,
+                                              exelist, LLVMDCompiler, full_version, for_machine,
                                               extra_args=[f])
             finally:
                 windows_proof_rm(f)
@@ -1152,7 +1158,7 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
                 exelist, version, for_machine, info, arch,
                 full_version=full_version, linker=linker, version_output=out)
         elif 'gdc' in out:
-            linker = guess_nix_linker(env, exelist, GnuDCompiler, for_machine)
+            linker = guess_nix_linker(env, exelist, GnuDCompiler, version, for_machine)
             return GnuDCompiler(
                 exelist, version, for_machine, info, arch,
                 exe_wrapper=exe_wrap, is_cross=is_cross,
@@ -1172,12 +1178,12 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
                 if info.is_windows() or info.is_cygwin():
                     objfile = os.path.basename(f)[:-1] + 'obj'
                     linker = guess_win_linker(env,
-                                              exelist, DmdDCompiler, for_machine,
+                                              exelist, DmdDCompiler, full_version, for_machine,
                                               invoked_directly=False, extra_args=[f, arch_arg])
                 else:
                     objfile = os.path.basename(f)[:-1] + 'o'
                     linker = guess_nix_linker(env,
-                                              exelist, DmdDCompiler, for_machine,
+                                              exelist, DmdDCompiler, full_version, for_machine,
                                               extra_args=[f, arch_arg])
             finally:
                 windows_proof_rm(f)
@@ -1186,7 +1192,7 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
             return DmdDCompiler(
                 exelist, version, for_machine, info, arch,
                 full_version=full_version, linker=linker)
-        raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
+        raise EnvironmentException('Unknown compiler: ' + join_args(exelist))
 
     _handle_exceptions(popen_exceptions, compilers)
     raise EnvironmentException('Unreachable code (exception to make mypy happy)')
@@ -1202,18 +1208,18 @@ def detect_swift_compiler(env: 'Environment', for_machine: MachineChoice) -> Com
     try:
         p, _, err = Popen_safe(exelist + ['-v'])
     except OSError:
-        raise EnvironmentException('Could not execute Swift compiler "{}"'.format(' '.join(exelist)))
+        raise EnvironmentException('Could not execute Swift compiler: {}'.format(join_args(exelist)))
     version = search_version(err)
     if 'Swift' in err:
         # As for 5.0.1 swiftc *requires* a file to check the linker:
         with tempfile.NamedTemporaryFile(suffix='.swift') as f:
             linker = guess_nix_linker(env,
-                                      exelist, SwiftCompiler, for_machine,
+                                      exelist, SwiftCompiler, version, for_machine,
                                       extra_args=[f.name])
         return SwiftCompiler(
             exelist, version, for_machine, is_cross, info, linker=linker)
 
-    raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
+    raise EnvironmentException('Unknown compiler: ' + join_args(exelist))
 
 
 # GNU/Clang defines and version
@@ -1229,9 +1235,12 @@ def _get_gnu_compiler_defines(compiler: T.List[str]) -> T.Dict[str, str]:
     if compiler[0].endswith('++'):
         args += ['-x', 'c++']
     args += ['-']
+    mlog.debug(f'Running command: {join_args(args)}')
     p, output, error = Popen_safe(args, write='', stdin=subprocess.PIPE)
     if p.returncode != 0:
-        raise EnvironmentException('Unable to detect GNU compiler type:\n' + output + error)
+        raise EnvironmentException('Unable to detect GNU compiler type:\n'
+                                   f'Compiler stdout:\n{output}\n-----\n'
+                                   f'Compiler stderr:\n{error}\n-----\n')
     # Parse several lines of the type:
     # `#define ___SOME_DEF some_value`
     # and extract `___SOME_DEF`
@@ -1256,9 +1265,12 @@ def _get_clang_compiler_defines(compiler: T.List[str]) -> T.Dict[str, str]:
     if compiler[0].endswith('++'):
         args += ['-x', 'c++']
     args += ['-']
+    mlog.debug(f'Running command: {join_args(args)}')
     p, output, error = Popen_safe(args, write='', stdin=subprocess.PIPE)
     if p.returncode != 0:
-        raise EnvironmentException('Unable to get clang pre-processor defines:\n' + output + error)
+        raise EnvironmentException('Unable to get clang pre-processor defines:\n'
+                                   f'Compiler stdout:\n{output}\n-----\n'
+                                   f'Compiler stderr:\n{error}\n-----\n')
     defines: T.Dict[str, str] = {}
     for line in output.split('\n'):
         if not line:
